#include <gmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "tmpfile.h"
#include "worker.h"

static char* progname;

void usage() {
   fprintf(stderr, "Usage: %s [-p N] N1 N2 {n_i}\n", progname);
   exit(1);
}

int main(int argc, char** argv) {
   progname = *argv++; --argc;
   unsigned int nof_processes = 4;
   if (argc > 1 && strcmp(*argv, "-p") == 0) {
      ++argv; --argc;
      int nof = atoi(*argv++); --argc;
      if (nof <= 0) usage();
      nof_processes = nof;
   }
   if (argc < 3) usage();

   mpz_t start; mpz_init_set_str(start, *argv++, 10); --argc;
   mpz_t end; mpz_init_set_str(end, *argv++, 10); --argc;

   int k = argc + 1;
   unsigned int offsets[k-1];
   for (int i = 0; i < k-1; ++i) {
      offsets[i] = atoi(*argv++); --argc;
   }

   /* len = end-start + 1 */
   mpz_t len; mpz_init(len);
   mpz_sub(len, end, start);
   mpz_add_ui(len, len, 1);

   /* FIXME: check that len is positive */

   /* q = len / nof_processes; r = len % nof_processes */
   mpz_t q; mpz_init(q); mpz_t r; mpz_init(r);
   mpz_tdiv_qr_ui(q, r, len, nof_processes);
   unsigned int remainder = mpz_get_ui(r);

   struct {
      pid_t pid;
      int outfd;
      mpz_t start_interval;
      mpz_t end_interval;
   } jobs[nof_processes];
   for (unsigned int i = 0; i < nof_processes; ++i) {
      int outfd = create_tmpfile();
      if (outfd < 0) perror("creating temporary file"), exit(1);
      jobs[i].outfd = outfd;

      /* start_interval = start + i * q;
         if (i < remainder) {
	    start_interval += i;
	 } else {
	    start_interval += remainder;
	 }
      */
      mpz_init_set(jobs[i].start_interval, start);
      mpz_addmul_ui(jobs[i].start_interval, q, i);
      if (i < remainder) {
	 mpz_add_ui(jobs[i].start_interval, jobs[i].start_interval, i);
      } else {
	 mpz_add_ui(jobs[i].start_interval, jobs[i].start_interval, remainder);
      }
   }
   for (unsigned int i = 0; i < nof_processes; ++i) {
      if (i == nof_processes - 1) {
	 mpz_init_set(jobs[i].end_interval, end);
      } else {
	 mpz_init(jobs[i].end_interval);
	 mpz_sub_ui(jobs[i].end_interval, jobs[i+1].start_interval, 1);
      }
   }

   /* fork ... */
   for (unsigned int i = 0; i < nof_processes; ++i) {
      pid_t pid = fork();
      if (pid < 0) perror("fork"), exit(1);
      if (pid == 0) {
	 find_primes(jobs[i].outfd, jobs[i].start_interval,
	    jobs[i].end_interval, k, offsets);
	 exit(0);
      }
      jobs[i].pid = pid;
   }
   /* ... and join */
   for (unsigned int i = 0; i < nof_processes; ++i) {
      int stat;
      if (waitpid(jobs[i].pid, &stat, 0) < 0) {
	 perror("wait"); exit(1);
      }
      int fd = jobs[i].outfd;
      if (lseek(fd, 0, SEEK_SET) < 0) {
	 perror("lseek"); exit(1);
      }
      char buf[BUFSIZ];
      ssize_t nbytes;
      while ((nbytes = read(fd, buf, sizeof buf)) > 0) {
	 write(1, buf, nbytes);
      }
      close(fd);
   }
}
